"""Support Muffin-Peewee."""
import peewee as pw
import wtforms as wtf
from muffin_peewee.models import to_simple

from muffin_rest import RESTHandler, Form, RESTNotFound, Filter, default_converter


try:
    from wtfpeewee.orm import model_form, ModelConverter

    ModelConverter.defaults[pw.DateField] = wtf.DateField
    ModelConverter.defaults[pw.DateTimeField] = wtf.DateTimeField

    class IntegerField(wtf.IntegerField):

        """Support null values."""

        def process_formdata(self, valuelist):
            """Support null values."""
            if valuelist:
                try:
                    self.data = int(valuelist[0])
                except (ValueError, TypeError):
                    self.raw_data = ['']
                    self.data = None
                    raise ValueError(self.gettext('Not a valid integer value'))

    ModelConverter.defaults[pw.IntegerField] = IntegerField

    from muffin_peewee.fields import JSONField

    ModelConverter.defaults[JSONField] = wtf.TextAreaField

except ImportError:
    model_form = None


def pw_converter(handler, flt):
    """Convert column name to filter."""
    return default_converter(handler, flt, PWFilter)


class PWRESTHandlerMeta(type(RESTHandler)):

    """Peewee specific."""

    def __new__(mcs, name, bases, params):
        """Prepare handler params."""
        model = params.get('model')
        params.setdefault('name', model and model._meta.db_table.lower() or name.lower())
        params.setdefault('model_pk', model and model._meta.primary_key or None)
        cls = super(PWRESTHandlerMeta, mcs).__new__(mcs, name, bases, params)
        if not cls.form and cls.model and model_form:
            cls.form = model_form(cls.model, base_class=Form, **cls.form_meta)
        return cls


class PWRESTHandler(RESTHandler, metaclass=PWRESTHandlerMeta):

    """Support REST for Peewee."""

    filters_converter = pw_converter

    model = None
    model_pk = None

    # only, exclude, recurse
    simple_meta = {}

    # only, exclude, field_args
    form_meta = {}

    def get_many(self, request):
        """Get collection."""
        return self.model.select()

    def get_one(self, request):
        """Load a resource."""
        resource = request.match_info.get(self.name)
        if not resource:
            return None

        try:
            return self.collection.where(self.model_pk == resource).get()
        except Exception:
            raise RESTNotFound(reason='Resource not found.')

    def populate(self):
        """Create object."""
        return self.model()

    def save_form(self, form, request, resource=None):
        """Save data."""
        resource = yield from super(PWRESTHandler, self).save_form(
            form, request, resource=resource)
        resource.save()
        return resource

    def to_simple(self, data, many=False):
        """Serialize data."""
        if many:
            return super(PWRESTHandler, self).to_simple(data, many)
        return to_simple(data, **self.simple_meta)

    def delete(self, request, resource=None):
        """Delete a resource."""
        if resource is None:
            raise RESTNotFound(reason='Resource not found')
        resource.delete_instance()

    def paginate(self, request, offset=0, limit=None):
        """Paginate queryset."""
        return self.collection.offset(offset).limit(limit), self.collection.count()


class PWFilter(Filter):

    """Base filter for Peewee handlers."""

    operations = dict(Filter.operations, **{
        '%': lambda a, b: a % b,
        '<<': lambda a, b: a << b,
    })

    def apply(self, query, value):
        """Filter a query."""
        form_field = query.model_class._meta.fields.get(self.column_name)
        if not form_field:
            return query
        return query.where(self.op(form_field, value))


# Filters for base primitives
PWBoolFilter = type('BoolFilter', (PWFilter,), {'form_field': wtf.BooleanField})
PWIntegerFilter = type('IntegerFilter', (PWFilter,), {'form_field': wtf.IntegerField})
PWDateFilter = type('DateFilter', (PWFilter,), {'form_field': wtf.DateField})
PWDateTimeFilter = type('DateTimeFilter', (PWFilter,), {'form_field': wtf.DateTimeField})


class PWMultiField(wtf.StringField):

    """Support many values."""

    def process_formdata(self, valuelist):
        """Save multivalues."""
        self.data = valuelist # noqa


class PWMultiFilter(PWFilter):

    """Support multi values."""

    form_field = PWMultiField

    def apply(self, query, value):
        """Filter a query."""
        form_field = query.model_class._meta.fields.get(self.column_name)
        if not form_field:
            return query

        return query.where(form_field << value)


class PWLikeFilter(PWFilter):

    """Filter query by value."""

    def apply(self, query, value):
        """Filter a query."""
        form_field = query.model_class._meta.fields.get(self.column_name)
        value = "*%s*" % value
        return query.where(form_field % value)



"""Handle REST."""
import datetime as dt
import math
from asyncio import iscoroutine
from urllib.parse import urlencode

from aiohttp import MultiDict
from aiohttp.web import StreamResponse, Response
from muffin.handler import Handler, abcoroutine
from ujson import dumps # noqa

from muffin_rest import RESTNotFound, RESTBadRequest, FILTER_PREFIX, default_converter, FilterForm


PAGE_VAR = FILTER_PREFIX + '-page'
LIMIT_VAR = FILTER_PREFIX + '-limit'


class RESTHandler(Handler):

    """Implement a common handler for REST operations."""

    #: A form for create/update the resource
    form = None

    #: Resource filters, it could be a field names or Filter instances
    filters = ()

    #: A function which convert field names to Filters
    filters_converter = default_converter

    #: If greater than zero collection of the resources will be paginated
    limit = 0

    def __init__(self):
        """Initialize a filters' form. And convert filters."""
        self.filters_form = FilterForm(prefix=FILTER_PREFIX)
        for flt in self.filters:
            field = self.filters_converter(flt)
            field.bind(self.filters_form)

    @classmethod
    def connect(cls, app, *paths, methods=None, name=None, **kwargs):
        """Connect to the application.

        Generate URL, name if it's not provided.
        """
        if not paths:
            paths = ['/%s(/{%s})?/?' % (cls.name, cls.name)]

        if name is None:
            name = "rest.%s" % cls.name

        return super(RESTHandler, cls).connect(app, *paths, methods=methods, name=name, **kwargs)

    @abcoroutine
    def dispatch(self, request, **kwgs):
        """Process request."""
        headers = {}

        # Authorization endpoint
        self.auth = yield from self.authorize(request)

        # Load collection
        self.collection = yield from self.get_many(request)

        if request.method == 'POST':
            return (yield from super(RESTHandler, self).dispatch(request, **kwgs))

        resource = yield from self.get_one(request)

        if request.method == 'GET' and resource is None:

            # Filter the collection
            if self.filters_form._fields:
                self.collection = yield from self.filter(request)

            # Paginate the collection
            try:
                limit = int(request.GET.get(LIMIT_VAR, self.limit))
                if limit:
                    curpage = int(request.GET.get(PAGE_VAR, 0))
                    offset = curpage * limit
                    self.collection, total = yield from self.paginate(request, offset, limit)
                    headers.update(make_pagination_headers(request, limit, curpage, total))
            except ValueError:
                raise RESTBadRequest('Pagination params are invalid.')

        response = yield from super(RESTHandler, self).dispatch(request, resource=resource, **kwgs)
        response.headers.update(headers)
        return response

    @abcoroutine
    def make_response(self, request, response):
        """Convert a handler result to web response."""
        while iscoroutine(response):
            response = yield from response

        if isinstance(response, StreamResponse):
            return response

        return Response(text=dumps(response), content_type='application/json')

    @abcoroutine
    def authorize(self, request):
        """Base point for authorization."""
        return True

    @abcoroutine
    def parse(self, request):
        """Ensure that request.data is multidict."""
        data = yield from super().parse(request)
        if not isinstance(data, MultiDict):
            data = MultiDict(data)
        return data

    @abcoroutine
    def get_many(self, request):
        """Base point for collect data."""
        return []

    @abcoroutine
    def filter(self, request):
        """Filter collection."""
        return self.filters_form.process(self.collection, request.GET)

    @abcoroutine
    def get_one(self, request):
        """Load resource."""
        return request.match_info.get(self.name)

    @abcoroutine
    def get_form(self, request, resource=None):
        """Initialize resource's form."""
        formdata = yield from self.parse(request)

        if not self.form:
            raise RESTBadRequest(reason="%s.form is not defined." % type(self).__name__)

        if resource:
            data = {}
            for name, field, *_ in self.form()._unbound_fields:
                value = getattr(resource, name, None)
                if value is None:
                    continue

                if isinstance(value, (dt.datetime, dt.date)):
                    field = field.bind(None, name, _meta=1)
                    value = value.strftime(field.format)
                data[name] = value

            data.update(formdata)
            formdata = MultiDict(data)

        return self.form(formdata, obj=resource)

    @abcoroutine
    def save_form(self, form, request, resource=None):
        """Save self form."""
        if resource is None:
            resource = self.populate()

        form.populate_obj(resource)
        return resource

    def populate(self):
        """Create a resource."""
        return object()

    def to_simple(self, data, many=False):
        """Serialize response to simple object (list, dict)."""
        if many:
            return [self.to_simple(r) for r in data]
        return str(data)

    @abcoroutine
    def get(self, request, resource=None):
        """Get resource or collection of resources."""
        if resource is not None and resource != '':
            return self.to_simple(resource)

        return self.to_simple(self.collection, many=True)

    @abcoroutine
    def post(self, request):
        """Create a resource."""
        form = yield from self.get_form(request)
        if not form.validate():
            raise RESTBadRequest(json=form.errors)
        resource = yield from self.save_form(form, request)
        return self.to_simple(resource)

    @abcoroutine
    def put(self, request, resource=None):
        """Update a resource."""
        if resource is None:
            raise RESTNotFound(reason='Resource not found')

        form = yield from self.get_form(request, resource=resource)
        if not form.validate():
            raise RESTBadRequest(json=form.errors)
        resource = yield from self.save_form(form, request, resource=resource)
        return self.to_simple(resource)

    patch = put

    @abcoroutine
    def batch(self, request):
        """Make group operations."""

    @abcoroutine
    def delete(self, request, resource=None):
        """Delete a resource."""
        if resource is None:
            raise RESTNotFound(reason='Resource not found')
        self.collection.remove(resource)

    @abcoroutine
    def paginate(self, request, offset=0, limit=0):
        """Paginate collection.

        :param request: client's request
        :param offset: current offset
        :param limit: limit items per page
        :returns: (paginated collection, count of resources)
        """
        return self.collection[offset: offset + limit], len(self.collection)

    @classmethod
    def scheme(cls):
        """Return self schema."""
        return {
            'form': cls.form and {
                name: str(type(field)) for (name, field) in cls.form()._fields.items()} or None,
            'methods': cls.methods,
            'description': cls.__doc__,
        }


def make_pagination_headers(request, limit, curpage, total):
    """Return Link Hypermedia Header."""
    lastpage = math.ceil(total / limit) - 1
    headers = {'X-Total-Count': str(total), 'X-Limit': str(limit),
               'X-Page-Last': str(lastpage), 'X-Page': str(curpage)}
    base = "{}?%s".format(request.path)
    links = {}
    links['first'] = base % urlencode(dict(request.GET, **{PAGE_VAR: 0}))
    links['last'] = base % urlencode(dict(request.GET, **{PAGE_VAR: lastpage}))
    if curpage:
        links['prev'] = base % urlencode(dict(request.GET, **{PAGE_VAR: curpage - 1}))
    if curpage < lastpage:
        links['next'] = base % urlencode(dict(request.GET, **{PAGE_VAR: curpage + 1}))

    headers['Link'] = ",".join(['<%s>; rel="%s"' % (v, n) for n, v in links.items()])
    return headers


#  pylama:ignore=W0201
